from __future__ import annotations

import logging

from django.conf import settings

from email_relay.conf import EMAIL_RELAY_DATABASE_ALIAS

pytest_plugins = []  # type: ignore


# Settings fixtures to bootstrap our tests
def pytest_configure(config):
    logging.disable(logging.CRITICAL)

    settings.configure(
        ALLOWED_HOSTS=["*"],
        CACHES={
            "default": {
                "BACKEND": "django.core.cache.backends.dummy.DummyCache",
            }
        },
        DATABASES={
            "default": {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": ":memory:",
            },
            EMAIL_RELAY_DATABASE_ALIAS: {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": ":memory:",
            },
        },
        DATABASE_ROUTERS=[
            "email_relay.db.EmailDatabaseRouter",
        ],
        EMAIL_BACKEND="django.core.mail.backends.locmem.EmailBackend",
        INSTALLED_APPS=[
            "django.contrib.contenttypes",
            "email_relay",
        ],
        LOGGING_CONFIG=None,
        PASSWORD_HASHERS=["django.contrib.auth.hashers.MD5PasswordHasher"],
        SECRET_KEY="NOTASECRET",
        USE_TZ=True,
    )



from __future__ import annotations

from email_relay.conf import app_settings


class EmailDatabaseRouter:
    def db_for_read(self, model, **hints):
        if model._meta.app_label == "email_relay":
            return app_settings.DATABASE_ALIAS
        return "default"

    def db_for_write(self, model, **hints):
        if model._meta.app_label == "email_relay":
            return app_settings.DATABASE_ALIAS
        return "default"

    def allow_relation(self, obj1, obj2, **hints):
        if (
            obj1._meta.app_label == "email_relay"
            or obj2._meta.app_label == "email_relay"
        ):
            return True
        return None

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if app_label == "email_relay":
            return db == app_settings.DATABASE_ALIAS
        return None



from __future__ import annotations

import datetime
import logging
from unittest import mock

import pytest
import responses
from django.core.management import call_command
from django.test.utils import override_settings
from django.utils import timezone
from model_bakery import baker

from email_relay.management.commands.runrelay import Command
from email_relay.models import Message
from email_relay.models import Status


def test_runrelay_help():
    # We'll capture the output of the command
    with pytest.raises(SystemExit) as exec_info:
        # call_command will execute our command as if we ran it from the command line
        # the 'stdout' argument captures the command output
        call_command("runrelay", "--help")

    # Asserting that the command exits with a successful exit code (0 for help command)
    assert exec_info.value.code == 0


@pytest.fixture
def runrelay():
    return Command()


@pytest.mark.django_db(databases=["default", "email_relay_db"])
def test_delete_sent_messages_based_on_retention_default(runrelay):
    baker.make(
        "email_relay.Message",
        status=Status.SENT,
        sent_at=timezone.now(),
        _quantity=10,
    )

    runrelay.delete_old_messages()

    assert Message.objects.count() == 10


@override_settings(
    DJANGO_EMAIL_RELAY={
        "MESSAGES_RETENTION_SECONDS": 0,
    }
)
@pytest.mark.django_db(databases=["default", "email_relay_db"])
def test_delete_sent_messages_based_on_retention_zero(runrelay):
    baker.make(
        "email_relay.Message",
        status=Status.SENT,
        sent_at=timezone.now(),
        _quantity=10,
    )

    runrelay.delete_old_messages()

    assert Message.objects.count() == 0


@override_settings(
    DJANGO_EMAIL_RELAY={
        "MESSAGES_RETENTION_SECONDS": 600,
    }
)
@pytest.mark.django_db(databases=["default", "email_relay_db"])
def test_delete_sent_messages_based_on_retention_non_zero(runrelay):
    baker.make(
        "email_relay.Message",
        status=Status.SENT,
        sent_at=timezone.now(),
        _quantity=5,
    )
    baker.make(
        "email_relay.Message",
        status=Status.SENT,
        sent_at=timezone.now() - datetime.timedelta(seconds=601),
        _quantity=5,
    )

    runrelay.delete_old_messages()

    assert Message.objects.count() == 5


@override_settings(
    DJANGO_EMAIL_RELAY={
        "RELAY_HEALTHCHECK_URL": "http://example.com/healthcheck",
    }
)
@responses.activate
def test_relay_healthcheck_url(runrelay, caplog):
    caplog.set_level(logging.DEBUG)
    responses.add(responses.GET, "http://example.com/healthcheck", status=200)

    runrelay.ping_healthcheck()

    assert len(responses.calls) == 1
    assert "healthcheck ping successful" in caplog.text


@responses.activate
def test_relay_healthcheck_url_not_configured(runrelay):
    responses.add(responses.GET, "http://example.com/healthcheck", status=200)

    runrelay.ping_healthcheck()

    assert len(responses.calls) == 0


@override_settings(
    DJANGO_EMAIL_RELAY={
        "RELAY_HEALTHCHECK_URL": "http://example.com/healthcheck",
        "RELAY_HEALTHCHECK_STATUS_CODE": 201,
    }
)
@responses.activate
def test_relay_healthcheck_status_code(runrelay, caplog):
    caplog.set_level(logging.DEBUG)
    responses.add(responses.GET, "http://example.com/healthcheck", status=201)

    runrelay.ping_healthcheck()

    assert len(responses.calls) == 1
    assert "healthcheck ping successful" in caplog.text


@override_settings(
    DJANGO_EMAIL_RELAY={
        "RELAY_HEALTHCHECK_URL": "http://example.com/healthcheck",
        "RELAY_HEALTHCHECK_METHOD": "POST",
    }
)
@responses.activate
def test_relay_healthcheck_method(runrelay, caplog):
    caplog.set_level(logging.DEBUG)
    responses.add(responses.POST, "http://example.com/healthcheck", status=200)

    runrelay.ping_healthcheck()

    assert len(responses.calls) == 1
    assert "healthcheck ping successful" in caplog.text


@override_settings(
    DJANGO_EMAIL_RELAY={
        "RELAY_HEALTHCHECK_URL": "http://example.com/healthcheck",
    }
)
@responses.activate
def test_relay_healthcheck_failure(runrelay, caplog):
    caplog.set_level(logging.WARNING)
    responses.add(responses.GET, "http://example.com/healthcheck", status=500)

    runrelay.ping_healthcheck()

    assert len(responses.calls) == 1
    assert "healthcheck ping successful" not in caplog.text


@override_settings(
    DJANGO_EMAIL_RELAY={
        "RELAY_HEALTHCHECK_URL": "http://example.com/healthcheck",
    }
)
def test_relay_healthcheck_no_requests(runrelay, caplog):
    caplog.set_level(logging.WARNING)

    with mock.patch("email_relay.management.commands.runrelay.requests", None):
        runrelay.ping_healthcheck()

    assert "Healthcheck URL configured but requests is not installed." in caplog.text



# Generated by Django 4.2.5 on 2023-09-27 18:38

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Message",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("data", models.JSONField()),
                (
                    "priority",
                    models.PositiveSmallIntegerField(
                        choices=[(1, "Low"), (2, "Medium"), (3, "High")], default=1
                    ),
                ),
                (
                    "status",
                    models.PositiveSmallIntegerField(
                        choices=[
                            (1, "Queued"),
                            (2, "Deferred"),
                            (3, "Failed"),
                            (4, "Sent"),
                        ],
                        default=1,
                    ),
                ),
                ("retry_count", models.PositiveSmallIntegerField(default=0)),
                (
                    "log",
                    models.TextField(
                        blank=True,
                        help_text="Most recent log message from the email backend, if any.",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("sent_at", models.DateTimeField(blank=True, null=True)),
            ],
            options={
                "ordering": ["created_at"],
            },
        ),
    ]



from __future__ import annotations

from collections.abc import Sequence

from django.core.mail import EmailMessage
from django.core.mail.backends.base import BaseEmailBackend

from email_relay.conf import app_settings
from email_relay.models import Message


class RelayDatabaseEmailBackend(BaseEmailBackend):
    def send_messages(self, email_messages: Sequence[EmailMessage]) -> int:
        messages = Message.objects.bulk_create(
            [Message(email=email) for email in email_messages],
            app_settings.MESSAGES_BATCH_SIZE,
        )
        return len(messages)



from __future__ import annotations

import logging
import smtplib
import time
from itertools import chain
from socket import error as socket_error

from django.conf import settings
from django.core.mail import get_connection
from django.db import transaction

from email_relay.conf import app_settings
from email_relay.models import Message

logger = logging.getLogger(__name__)


def send_all():
    logger.info("sending emails")

    counts = {
        "sent": 0,
        "deferred": 0,
    }

    message_batch = list(
        chain(
            Message.objects.queued().prioritized(),
            Message.objects.deferred().prioritized(),
        )
    )
    logger.debug(f"found {len(message_batch)} messages to send")

    if app_settings.EMAIL_MAX_BATCH is not None:
        msg = f"max batch size is {app_settings.EMAIL_MAX_BATCH}"
        if len(message_batch) > app_settings.EMAIL_MAX_BATCH:
            msg += ", truncating"
        logger.debug(msg)
        message_batch = message_batch[: app_settings.EMAIL_MAX_BATCH]

    connection = None

    for message in message_batch:
        with transaction.atomic():
            try:
                message = (
                    Message.objects.filter(id=message.id)
                    .select_for_update(skip_locked=True)
                    .get()
                )
            except Message.DoesNotExist:
                continue
            try:
                if connection is None:
                    relay_email_backend = getattr(
                        settings,
                        "EMAIL_BACKEND",
                        "django.core.mail.backends.smtp.EmailBackend",
                    )
                    connection = get_connection(backend=relay_email_backend)
                email = message.email
                if email is not None:
                    email.connection = connection
                    email.send()
                    logger.debug(f"sent message {message.id}")
                    message.mark_sent()
                    counts["sent"] += 1
                else:
                    msg = f"Message {message.id} has no email object"
                    message.fail(log=msg)
                    logger.warning(msg)
            except Exception as err:
                handled_exceptions = (
                    smtplib.SMTPAuthenticationError,
                    smtplib.SMTPDataError,
                    smtplib.SMTPRecipientsRefused,
                    smtplib.SMTPSenderRefused,
                    socket_error,
                )
                if isinstance(err, handled_exceptions):
                    logger.debug(f"deferring message {message.id} due to {err}")
                    message.defer(log=str(err))
                    if message.retry_count >= app_settings.EMAIL_MAX_RETRIES:
                        logger.warning(
                            f"max retries reached, marking message {message.id} as failed"
                        )
                        message.fail(log=str(err))
                    connection = None
                    counts["deferred"] += 1
                else:
                    raise err

        if (
            app_settings.EMAIL_MAX_DEFERRED is not None
            and counts["deferred"] >= app_settings.EMAIL_MAX_DEFERRED
        ):
            logger.debug(
                f"max deferred emails reached ({app_settings.EMAIL_MAX_DEFERRED}), stopping"
            )
            break

        if app_settings.EMAIL_THROTTLE > 0:
            logger.debug(
                f"throttling enabled, sleeping for {app_settings.EMAIL_THROTTLE} seconds"
            )
            time.sleep(app_settings.EMAIL_THROTTLE)

    logger.info(f"sent {counts['sent']} emails, deferred {counts['deferred']} emails")



from __future__ import annotations

import os

from service.utils import coerce_dict_values
from service.utils import env_vars_to_nested_dict
from service.utils import filter_valid_django_settings
from service.utils import get_user_settings_from_env
from service.utils import merge_with_defaults


def test_env_vars_to_nested_dict():
    env_vars = {
        "DATABASES__default__CONN_MAX_AGE": 600,
        "DEBUG": "True",
    }

    assert env_vars_to_nested_dict(env_vars) == {
        "DATABASES": {
            "default": {
                "CONN_MAX_AGE": 600,
            }
        },
        "DEBUG": "True",
    }


def test_merge_with_defaults():
    default_settings = {
        "DATABASES": {
            "default": {
                "CONN_MAX_AGE": 600,
            }
        },
        "DEBUG": False,
    }
    user_settings = {
        "DATABASES": {
            "default": {
                "CONN_MAX_AGE": 300,
            }
        },
        "DEBUG": True,
    }

    assert merge_with_defaults(default_settings, user_settings) == {
        "DATABASES": {
            "default": {
                "CONN_MAX_AGE": 300,
            }
        },
        "DEBUG": True,
    }


def test_get_user_settings_from_env():
    env_vars = {
        "DATABASES__default__CONN_MAX_AGE": "600",
        "DEBUG": "True",
        "INVALID_KEY": "True",
    }
    for k, v in env_vars.items():
        os.environ[k] = v

    assert get_user_settings_from_env() == {
        "DATABASES": {
            "default": {
                "CONN_MAX_AGE": 600,
            }
        },
        "DEBUG": True,
    }

    for k in env_vars.keys():
        del os.environ[k]


def test_coerce_dict_values():
    types_dict = {
        "BOOLEAN": "True",
        "INTEGER": "600",
        "STRING": "str",
        "FLOAT": "3.14",
        "NONE": "None",
    }

    d = {
        **types_dict,
        "NESTED": types_dict,
    }

    assert coerce_dict_values(d) == {
        "BOOLEAN": True,
        "INTEGER": 600,
        "STRING": "str",
        "FLOAT": 3.14,
        "NONE": None,
        "NESTED": {
            "BOOLEAN": True,
            "INTEGER": 600,
            "STRING": "str",
            "FLOAT": 3.14,
            "NONE": None,
        },
    }


def test_filter_valid_django_settings():
    d = {
        "DEBUG": True,
        "INVALID_KEY": "invalid value",
        "DJANGO_EMAIL_RELAY": {
            "VALID_KEY": "valid value",
        },
    }

    assert filter_valid_django_settings(d) == {
        "DEBUG": True,
        "DJANGO_EMAIL_RELAY": {
            "VALID_KEY": "valid value",
        },
    }



from __future__ import annotations

import base64
import binascii
import datetime
from email.mime.base import MIMEBase

from django.core.mail import EmailMessage
from django.core.mail import EmailMultiAlternatives
from django.db import models
from django.utils import timezone


class Priority(models.IntegerChoices):
    LOW = 1, "Low"
    MEDIUM = 2, "Medium"
    HIGH = 3, "High"


class Status(models.IntegerChoices):
    QUEUED = 1, "Queued"
    DEFERRED = 2, "Deferred"
    FAILED = 3, "Failed"
    SENT = 4, "Sent"


class MessageQuerySet(models.QuerySet):
    def prioritized(self):
        return self.order_by("-priority", "created_at")

    def high_priority(self):
        return self.filter(priority=Priority.HIGH)

    def medium_priority(self):
        return self.filter(priority=Priority.MEDIUM)

    def low_priority(self):
        return self.filter(priority=Priority.LOW)

    def queued(self):
        return self.filter(status=Status.QUEUED)

    def deferred(self):
        return self.filter(status=Status.DEFERRED)

    def failed(self):
        return self.filter(status=Status.FAILED)

    def sent(self):
        return self.filter(status=Status.SENT)

    def sent_before(self, dt: datetime.datetime):
        return self.sent().filter(sent_at__lte=dt)


class Message(models.Model):
    data = models.JSONField()
    priority = models.PositiveSmallIntegerField(
        choices=Priority.choices, default=Priority.LOW
    )
    status = models.PositiveSmallIntegerField(
        choices=Status.choices, default=Status.QUEUED
    )
    retry_count = models.PositiveSmallIntegerField(default=0)
    log = models.TextField(
        blank=True, help_text="Most recent log message from the email backend, if any."
    )

    created_at = models.DateTimeField(auto_now_add=True, editable=False)
    updated_at = models.DateTimeField(auto_now=True, editable=False)
    sent_at = models.DateTimeField(null=True, blank=True)

    objects = MessageQuerySet.as_manager()

    class Meta:
        ordering = ["created_at"]

    def __str__(self):
        try:
            return f'{self.created_at} "{self.data["subject"]}" to {", ".join(self.data["to"])}'
        except Exception:
            return f"{self.created_at} <invalid message>"

    def save(self, *args, **kwargs):
        # Overriding the save method in order to make sure that
        # modified field is updated even if it is not given as
        # a parameter to the update field argument.
        update_fields = kwargs.get("update_fields", None)
        if update_fields:
            kwargs["update_fields"] = set(update_fields).union({"updated_at"})

        super().save(*args, **kwargs)

    def mark_sent(self):
        self.status = Status.SENT
        self.sent_at = timezone.now()
        self.save()

    def defer(self, log: str = ""):
        self.status = Status.DEFERRED
        self.log = log
        self.retry_count += 1
        self.save()

    def fail(self, log: str = ""):
        self.status = Status.FAILED
        self.log = log
        self.save()

    @property
    def email(self) -> EmailMessage | None:
        data = self.data
        if not data:
            return None

        email = EmailMultiAlternatives(
            subject=data.get("subject", ""),
            body=data.get("message"),
            from_email=data.get("from_email"),
            to=data.get("recipient_list"),
        )

        html_message = data.get("html_message", None)
        if html_message:
            email.attach_alternative(html_message, "text/html")

        for attachment in data.get("attachments", []):
            if isinstance(attachment, dict):
                content = attachment.get("content")
                if isinstance(content, str):
                    try:
                        # Attempt to decode the base64 string into bytes
                        decoded_content = base64.b64decode(content)
                    except binascii.Error:
                        # Fallback to assuming it's plain text, encoded as bytes
                        decoded_content = content.encode("utf-8")

                    email.attach(
                        filename=attachment.get("filename", ""),
                        content=decoded_content,
                        mimetype=attachment.get("mimetype", ""),
                    )
                else:
                    email.attach(**attachment)
            else:
                email.attach(*attachment)

        return email

    @email.setter
    def email(self, email_message: EmailMessage) -> None:
        self.data = {
            "subject": email_message.subject,
            "message": email_message.body,
            "from_email": email_message.from_email,
            "recipient_list": email_message.to,
            "html_message": next(
                (
                    alternative[0]
                    for alternative in getattr(email_message, "alternatives", [])
                    if alternative[1] == "text/html"
                ),
                None,
            ),
            "attachments": [
                {
                    "filename": attachment[0],
                    "content": base64.b64encode(attachment[1]).decode("utf-8")
                    if isinstance(attachment[1], bytes)
                    else attachment[1],
                    "mimetype": attachment[2],
                }
                if not isinstance(attachment, MIMEBase)
                else attachment
                for attachment in email_message.attachments
            ],
        }



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from django.conf import settings

EMAIL_RELAY_SETTINGS_NAME = "DJANGO_EMAIL_RELAY"
EMAIL_RELAY_DATABASE_ALIAS = "email_relay_db"


@dataclass
class AppSettings:
    DATABASE_ALIAS: str = EMAIL_RELAY_DATABASE_ALIAS
    EMAIL_MAX_BATCH: int | None = None
    EMAIL_MAX_DEFERRED: int | None = None
    EMAIL_MAX_RETRIES: int | None = None
    EMPTY_QUEUE_SLEEP: int = 30
    EMAIL_THROTTLE: int = 0
    MESSAGES_BATCH_SIZE: int | None = None
    MESSAGES_RETENTION_SECONDS: int | None = None
    RELAY_HEALTHCHECK_METHOD: str = "GET"
    RELAY_HEALTHCHECK_STATUS_CODE: int = 200
    RELAY_HEALTHCHECK_TIMEOUT: float | tuple[float, float] | tuple[float, None] = 5.0
    RELAY_HEALTHCHECK_URL: str | None = None

    def __getattribute__(self, __name: str) -> Any:
        user_settings = getattr(settings, EMAIL_RELAY_SETTINGS_NAME, {})
        return user_settings.get(__name, super().__getattribute__(__name))


app_settings = AppSettings()



from __future__ import annotations

import os
from pathlib import Path

import nox

PY38 = "3.8"
PY39 = "3.9"
PY310 = "3.10"
PY311 = "3.11"
PY312 = "3.12"
PY_VERSIONS = [PY38, PY39, PY310, PY311, PY312]
PY_DEFAULT = PY38

DJ32 = "3.2"
DJ41 = "4.1"
DJ42 = "4.2"
DJMAIN = "main"
DJMAIN_MIN_PY = PY310
DJ_VERSIONS = [DJ32, DJ41, DJ42, DJMAIN]
DJ_DEFAULT = DJ32


def version(ver: str) -> tuple[int, ...]:
    """Convert a string version to a tuple of ints, e.g. "3.10" -> (3, 10)"""
    return tuple(map(int, ver.split(".")))


def should_skip(python: str, django: str) -> tuple[bool, str | None]:
    """Return True if the test should be skipped"""
    if django == DJMAIN and version(python) < version(DJMAIN_MIN_PY):
        return True, f"Django {DJMAIN} requires Python {DJMAIN_MIN_PY}+"

    if django == DJ32 and version(python) >= version(PY312):
        return True, f"Django {DJ32} requires Python < {PY312}"

    return False, None


@nox.session(python=PY_VERSIONS)
@nox.parametrize("django", DJ_VERSIONS)
def tests(session, django):
    skip = should_skip(session.python, django)
    if skip[0]:
        session.skip(skip[1])

    session.install(".[dev,relay]")

    if django == DJMAIN:
        session.install("https://github.com/django/django/archive/refs/heads/main.zip")
    else:
        session.install(f"django=={django}")

    session.run("python", "-m", "pytest")


@nox.session
def coverage(session):
    session.install(".[dev,relay]")
    session.run("python", "-m", "pytest", "--cov=email_relay", "--cov=service")

    try:
        summary = os.environ["GITHUB_STEP_SUMMARY"]
        with Path(summary).open("a") as output_buffer:
            output_buffer.write("")
            output_buffer.write("### Coverage\n\n")
            output_buffer.flush()
            session.run(
                "python",
                "-m",
                "coverage",
                "report",
                "--skip-covered",
                "--skip-empty",
                "--format=markdown",
                stdout=output_buffer,
            )
    except KeyError:
        session.run(
            "python", "-m", "coverage", "html", "--skip-covered", "--skip-empty"
        )

    session.run("python", "-m", "coverage", "report")


@nox.session
def lint(session):
    session.install(".[lint]")
    session.run("python", "-m", "pre_commit", "run", "--all-files")


@nox.session
def mypy(session):
    session.install(".[dev,relay]")
    session.run("python", "-m", "mypy", ".")



from __future__ import annotations

import pytest

from email_relay.conf import app_settings
from email_relay.db import EmailDatabaseRouter


# Mock model with app_label "email_relay"
class MockModel:
    class _meta:
        app_label = "email_relay"


# Mock model with app_label "some_other_app"
class MockModelOther:
    class _meta:
        app_label = "some_other_app"


@pytest.fixture
def router():
    return EmailDatabaseRouter()


def test_db_for_read(router):
    assert router.db_for_read(MockModel) == app_settings.DATABASE_ALIAS
    assert router.db_for_read(MockModelOther) == "default"


def test_db_for_write(router):
    assert router.db_for_write(MockModel) == app_settings.DATABASE_ALIAS
    assert router.db_for_write(MockModelOther) == "default"


def test_allow_relation(router):
    assert router.allow_relation(MockModel, MockModel)


def test_allow_migrate(router):
    assert router.allow_migrate(app_settings.DATABASE_ALIAS, "email_relay")



from __future__ import annotations

from django.conf import settings
from django.test import override_settings

from email_relay.conf import app_settings


def test_default_settings():
    user_settings = getattr(settings, "DJANGO_EMAIL_RELAY", {})

    assert user_settings == {}
    assert app_settings.DATABASE_ALIAS == "email_relay_db"
    assert app_settings.MESSAGES_BATCH_SIZE is None
    assert app_settings.EMAIL_MAX_RETRIES is None


@override_settings(
    DJANGO_EMAIL_RELAY={
        "DATABASE_ALIAS": "custom_db_name",
    },
)
def test_custom_database_alias():
    assert app_settings.DATABASE_ALIAS == "custom_db_name"


@override_settings(
    DJANGO_EMAIL_RELAY={
        "MESSAGES_BATCH_SIZE": 10,
    },
)
def test_custom_batch_size():
    assert app_settings.MESSAGES_BATCH_SIZE == 10


@override_settings(
    DJANGO_EMAIL_RELAY={
        "EMAIL_MAX_RETRIES": 5,
    },
)
def test_custom_max_retries():
    assert app_settings.EMAIL_MAX_RETRIES == 5



from __future__ import annotations

import pytest
from django.conf import settings
from django.core.mail import EmailMessage
from django.core.mail import send_mail
from django.test.utils import override_settings

from email_relay.models import Message


@pytest.fixture(scope="module", autouse=True)
def relay_backend():
    with override_settings(
        EMAIL_BACKEND="email_relay.backend.RelayDatabaseEmailBackend"
    ):
        yield


def test_fixture():
    assert settings.EMAIL_BACKEND == "email_relay.backend.RelayDatabaseEmailBackend"


@pytest.mark.django_db(databases=["default", "email_relay_db"])
def test_send_mail():
    assert Message.objects.count() == 0

    send_mail(
        "Subject here",
        "Here is the message.",
        "from_test@example.com",
        ["to_test@example.com"],
    )

    assert Message.objects.count() == 1


@pytest.mark.django_db(databases=["default", "email_relay_db"])
def test_email_message():
    assert Message.objects.count() == 0

    email = EmailMessage(
        "Subject here",
        "Here is the message.",
        "from_test@example.com",
        ["to_test@example.com"],
    )

    email.send()

    assert Message.objects.count() == 1



from __future__ import annotations

import datetime
import base64

import pytest
from django.utils import timezone
from django.core.mail import EmailMessage, EmailMultiAlternatives
from model_bakery import baker

from email_relay.models import Message
from email_relay.models import Priority
from email_relay.models import Status


@pytest.mark.django_db(databases=["default", "email_relay_db"])
def test_message():
    baker.make("email_relay.Message")
    assert Message.objects.all().count() == 1


@pytest.mark.django_db(databases=["default", "email_relay_db"])
class TestMessageQuerySet:
    @pytest.fixture
    def messages_with_priority(self):
        low = baker.make("email_relay.Message", priority=Priority.LOW)
        medium = baker.make("email_relay.Message", priority=Priority.MEDIUM)
        high = baker.make("email_relay.Message", priority=Priority.HIGH)
        return {
            "low": low,
            "medium": medium,
            "high": high,
        }

    @pytest.fixture
    def messages_with_status(self):
        queued = baker.make("email_relay.Message", status=Status.QUEUED)
        deferred = baker.make("email_relay.Message", status=Status.DEFERRED)
        failed = baker.make("email_relay.Message", status=Status.FAILED)
        sent = baker.make("email_relay.Message", status=Status.SENT)
        return {
            "queued": queued,
            "deferred": deferred,
            "failed": failed,
            "sent": sent,
        }

    def test_prioritized(self, messages_with_priority):
        queryset = Message.objects.prioritized()

        assert queryset.count() == 3
        assert queryset[0] == messages_with_priority["high"]
        assert queryset[1] == messages_with_priority["medium"]
        assert queryset[2] == messages_with_priority["low"]

    def test_high_priority(self, messages_with_priority):
        queryset = Message.objects.high_priority()

        assert queryset.count() == 1
        assert queryset[0] == messages_with_priority["high"]

    def test_medium_priority(self, messages_with_priority):
        queryset = Message.objects.medium_priority()

        assert queryset.count() == 1
        assert queryset[0] == messages_with_priority["medium"]

    def test_low_priority(self, messages_with_priority):
        queryset = Message.objects.low_priority()

        assert queryset.count() == 1
        assert queryset[0] == messages_with_priority["low"]

    def test_queued(self, messages_with_status):
        queryset = Message.objects.queued()

        assert queryset.count() == 1
        assert queryset[0] == messages_with_status["queued"]

    def test_deferred(self, messages_with_status):
        queryset = Message.objects.deferred()

        assert queryset.count() == 1
        assert queryset[0] == messages_with_status["deferred"]

    def test_failed(self, messages_with_status):
        queryset = Message.objects.failed()

        assert queryset.count() == 1
        assert queryset[0] == messages_with_status["failed"]

    def test_sent(self, messages_with_status):
        queryset = Message.objects.sent()

        assert queryset.count() == 1
        assert queryset[0] == messages_with_status["sent"]

    def test_sent_before(self):
        one_week = baker.make(
            "email_relay.Message",
            status=Status.SENT,
            sent_at=timezone.now() - datetime.timedelta(days=7),
        )
        now = baker.make(
            "email_relay.Message",
            status=Status.SENT,
            sent_at=timezone.now(),
        )
        not_sent = baker.make(
            "email_relay.Message",
            status=Status.QUEUED,
            sent_at=None,
        )

        queryset = Message.objects.sent_before(
            timezone.now() - datetime.timedelta(days=1)
        )

        assert queryset.count() == 1
        assert one_week in queryset
        assert now not in queryset
        assert not_sent not in queryset


@pytest.mark.django_db(databases=["default", "email_relay_db"])
class TestMessageModel:
    @pytest.fixture
    def data(self):
        return {
            "subject": "Test",
            "message": "Test",
            "from_email": "from@example.com",
            "recipient_list": ["to@example.com"],
        }

    @pytest.fixture
    def email(self):
        return EmailMultiAlternatives(
            subject="Test",
            body="Test",
            from_email="from@example.com",
            to=["to@example.com"],
        )

    def test_create_message(self, data):
        message = Message.objects.create(data=data)

        assert message.data == data
        assert message.priority == Priority.LOW
        assert message.status == Status.QUEUED
        assert message.retry_count == 0
        assert message.log == ""
        assert message.sent_at is None

    def test_email_property(self, data):
        message = Message.objects.create(data=data)

        email = message.email

        assert isinstance(email, EmailMessage)
        assert email.subject == data["subject"]
        assert email.body == data["message"]
        assert email.from_email == data["from_email"]
        assert email.to == data["recipient_list"]

    def test_email_setter(self, data):
        message = Message.objects.create(data=data)
        email = EmailMultiAlternatives(
            subject="Test 2",
            body="Test 2",
            from_email="from2@example.com",
            to=["to2@example.com"],
        )

        message.email = email
        message.save()

        assert message.data["subject"] == email.subject
        assert message.data["message"] == email.body
        assert message.data["from_email"] == email.from_email
        assert message.data["recipient_list"] == email.to

    def test_email_with_plain_text_attachment(self, email):
        attachment_content = b"Hello World!"
        email.attach(
            filename="test.txt",
            content=attachment_content,
            mimetype="text/plain",
        )

        message = Message()
        message.email = email
        message.save()

        assert Message.objects.count() == 1

        saved_message = Message.objects.first()
        assert saved_message.data["attachments"][0]["filename"] == "test.txt"
        assert saved_message.data["attachments"][0][
            "content"
        ] == attachment_content.decode("utf-8")
        assert saved_message.data["attachments"][0]["mimetype"] == "text/plain"

        email_from_db = saved_message.email
        assert email_from_db.attachments[0][0] == "test.txt"
        assert email_from_db.attachments[0][1] == attachment_content.decode("utf-8")
        assert email_from_db.attachments[0][2] == "text/plain"

    def test_email_with_binary_attachment(self, email, faker):
        attachment_content = faker.binary(length=10)
        email.attach(
            filename="test.zip",
            content=attachment_content,
            mimetype="application/zip",
        )

        message = Message()
        message.email = email
        message.save()

        assert Message.objects.count() == 1

        saved_message = Message.objects.first()
        assert saved_message.data["attachments"][0]["filename"] == "test.zip"
        assert saved_message.data["attachments"][0]["content"] == base64.b64encode(
            attachment_content
        ).decode("utf-8")
        assert saved_message.data["attachments"][0]["mimetype"] == "application/zip"

        email_from_db = saved_message.email
        assert email_from_db.attachments[0][0] == "test.zip"
        assert email_from_db.attachments[0][1] == attachment_content
        assert email_from_db.attachments[0][2] == "application/zip"

    def test_email_send(self, email, mailoutbox):
        message = Message()
        message.email = email
        message.save()

        message.email.send()

        assert len(mailoutbox) == 1

    def test_email_send_with_plain_text_attachment(self, email, mailoutbox):
        email.attach(
            filename="test.txt",
            content=b"Hello World!",
            mimetype="text/plain",
        )
        message = Message()
        message.email = email
        message.save()

        message.email.send()

        assert len(mailoutbox) == 1

    def test_email_send_with_binary_attachment(self, email, faker, mailoutbox):
        email.attach(
            filename="test.zip",
            content=faker.binary(length=10),
            mimetype="application/zip",
        )
        message = Message()
        message.email = email
        message.save()

        message.email.send()

        assert len(mailoutbox) == 1



from __future__ import annotations

import datetime
import logging
import time

from django.core.management import BaseCommand
from django.utils import timezone

from email_relay.conf import app_settings
from email_relay.models import Message
from email_relay.relay import send_all

try:
    import requests
except ImportError:
    requests = None  # type: ignore[assignment]

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def handle(self, *args, **options) -> None:
        logger.info("starting relay")
        while True:
            while (
                not Message.objects.queued().exists()
                and not Message.objects.deferred().exists()
            ):
                msg = "queue is empty"
                if app_settings.EMPTY_QUEUE_SLEEP > 0:
                    msg += f", sleeping for {app_settings.EMPTY_QUEUE_SLEEP} seconds before checking queue again"
                    time.sleep(app_settings.EMPTY_QUEUE_SLEEP)
                logger.debug(msg)

            send_all()
            self.delete_old_messages()
            self.ping_healthcheck()

    def delete_old_messages(self) -> None:
        if app_settings.MESSAGES_RETENTION_SECONDS is not None:
            logger.debug("deleting old messages")
            if app_settings.MESSAGES_RETENTION_SECONDS == 0:
                deleted_messages = Message.objects.sent().delete()
            else:
                deleted_messages = Message.objects.sent_before(
                    timezone.now()
                    - datetime.timedelta(
                        seconds=app_settings.MESSAGES_RETENTION_SECONDS
                    )
                ).delete()
            logger.debug(f"deleted {deleted_messages[0]} messages")

    def ping_healthcheck(self) -> None:
        if app_settings.RELAY_HEALTHCHECK_URL is not None:
            if requests is None:
                logger.warning(
                    "Healthcheck URL configured but requests is not installed. "
                    "Please install requests to use the healthcheck feature."
                )
                return

            response = requests.request(
                method=app_settings.RELAY_HEALTHCHECK_METHOD,
                url=app_settings.RELAY_HEALTHCHECK_URL,
                timeout=app_settings.RELAY_HEALTHCHECK_TIMEOUT,
            )

            if response.status_code == app_settings.RELAY_HEALTHCHECK_STATUS_CODE:
                logger.debug("healthcheck ping successful")
            else:
                logger.warning(
                    f"healthcheck failed, got {response.status_code}, expected {app_settings.RELAY_HEALTHCHECK_STATUS_CODE}"
                )
